#include "verto.h"
#include "verto-jsonrpc.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>

bool
subtract(verto_ctx *ctx, const char *method, json_object *params,
         void *misc, json_object **response)
{
  int a, b;
  json_object *x, *y;

  if (!response)
    return true;

  if (json_object_is_type(params, json_type_array)) {
    x = json_object_array_get_idx(params, 0);
    y = json_object_array_get_idx(params, 1);
  } else if (json_object_is_type(params, json_type_object)) {
    x = json_object_object_get(params, "minuend");
    y = json_object_object_get(params, "subtrahend");
  } else {
    *response = json_object_new_object();
    if (!*response)
      return false;

    x = json_object_new_int(JSONRPC_ERRCODE_INVALID_PARAMS);
    y = json_object_new_string("Invalid params");
    if (!x || !y) {
      json_object_put(*response);
      json_object_put(x);
      json_object_put(y);
      *response = NULL;
      return false;
    }

    json_object_object_add(*response, "code", x);
    json_object_object_add(*response, "message", y);
    return true;
  }

  assert(json_object_get_type(x) == json_type_int);
  assert(json_object_get_type(y) == json_type_int);

  a = json_object_get_int(x);
  b = json_object_get_int(y);

  *response = json_object_new_int(a - b);
  return *response != NULL;
}

bool
update_sum(verto_ctx *ctx, const char *method, json_object *params,
           void *misc, json_object **response)
{
  json_object *x, *y;
  int i, len, sum = 0;

  if (!json_object_is_type(params, json_type_array))
    goto invalid_params;

  len = json_object_array_length(params);
  for (i=0 ; i < len ; i++) {
    x = json_object_array_get_idx(params, i);
    if (!x || !json_object_is_type(x, json_type_int))
      goto invalid_params;

    sum += json_object_get_int(x);
  }

  if (misc)
    *((int*) misc) = sum;
  if (response) {
    *response = json_object_new_int(sum);
    return *response != NULL;
  } else
    return true;

invalid_params:
  *response = json_object_new_object();
  if (!*response)
    return false;

  x = json_object_new_int(JSONRPC_ERRCODE_INVALID_PARAMS);
  y = json_object_new_string("Invalid params");
  if (!x || !y) {
    json_object_put(*response);
    json_object_put(x);
    json_object_put(y);
    *response = NULL;
    return false;
  }

  json_object_object_add(*response, "code", x);
  json_object_object_add(*response, "message", y);
  return true;
}

bool
get_data(verto_ctx *ctx, const char *method, json_object *params,
         void *misc, json_object **response)
{
  if (response)
    *response = json_tokener_parse("[\"hello\", 5]");
  return response ? *response ? true : false : true;
}

static int subtract_count = 0;
static int foobar_count = 0;
static int sum_count = 0;
static int get_data_count = 0;

void
subtract_cb(json_object *response, bool error, void *misc)
{
  assert(json_object_get_int(response) == ((int*) misc)[subtract_count]);
  printf("SUBTRACT %d:\tOK\n", ++subtract_count);
}

void
foobar_cb(json_object *response, bool error, void *misc)
{
  json_object *x;

  assert(error);
  x = json_object_object_get(response, "code");
  assert(json_object_is_type(x, json_type_int));
  assert(json_object_get_int(x) == JSONRPC_ERRCODE_METHOD_NOT_FOUND);
  printf("FOOBAR %d:\tOK\n", ++foobar_count);
}

void
sum_cb(json_object *response, bool error, void *misc)
{
  assert(json_object_get_int(response) == ((int*) misc)[sum_count]);
  printf("SUM %d:\t\tOK\n", ++sum_count);
}

void
get_data_cb(json_object *response, bool error, void *misc)
{
  json_object *x;
  assert(!error);
  assert(json_object_is_type(response, json_type_array));
  x = json_object_array_get_idx(response, 0);
  assert(x);
  assert(json_object_is_type(x, json_type_string));
  assert(!strcmp(json_object_get_string(x), "hello"));
  x = json_object_array_get_idx(response, 1);
  assert(x);
  assert(json_object_is_type(x, json_type_int));
  assert(json_object_get_int(x) == 5);
  get_data_count++;
}

void
on_idle(verto_ctx *ctx, verto_ev *ev)
{
  verto_break(ctx);
}

void
err_cb(verto_ev *ev, int errno, void *misc)
{
  assert(false);
}

int
main(int argc, const char **argv)
{
  verto_ctx *ctx = NULL;
  verto_ev *client = NULL, *server = NULL;
  jsonrpc_batch *batch;
  int sockpair[2] = { -1, -1 };
  int retval = 0;

  ctx = verto_new(NULL, VERTO_EV_TYPE_IO | VERTO_EV_TYPE_IDLE);
  assert(ctx);
  assert(socketpair(AF_UNIX, SOCK_STREAM, 0, sockpair) == 0);

  client = jsonrpc_init(ctx, sockpair[0], sockpair[0], 8 * 1024, err_cb, NULL, NULL);
  server = jsonrpc_init(ctx, sockpair[1], sockpair[1], 8 * 1024, err_cb, NULL, NULL);
  assert(client && server);

  int subtract_rslts[] = {19, -19, 19, 19, 19};
  int update_rslt = 0;
  assert(jsonrpc_set_handler(server, "subtract", subtract, NULL, NULL));
  assert(jsonrpc_set_handler(server, "update",   update_sum, &update_rslt, NULL));
  assert(jsonrpc_set_handler(server, "sum",      update_sum, NULL, NULL));
  assert(jsonrpc_set_handler(server, "get_data", get_data, NULL, NULL));

  assert(jsonrpc_call(client, NULL, "subtract", json_tokener_parse("[ 42, 23 ]"), subtract_cb, subtract_rslts, NULL));
  assert(jsonrpc_call(client, NULL, "subtract", json_tokener_parse("[ 23, 42 ]"), subtract_cb, subtract_rslts, NULL));
  assert(jsonrpc_call(client, NULL, "subtract", json_tokener_parse("{\"subtrahend\": 23, \"minuend\": 42}"), subtract_cb, subtract_rslts, NULL));
  assert(jsonrpc_call(client, NULL, "subtract", json_tokener_parse("{\"minuend\": 42, \"subtrahend\": 23}"), subtract_cb, subtract_rslts, NULL));
  assert(jsonrpc_call(client, NULL, "update", json_tokener_parse("[ 1, 2, 3, 4, 5 ]"), NULL, NULL, NULL));
  assert(jsonrpc_call(client, NULL, "foobar", json_tokener_parse("[]"), foobar_cb, NULL, NULL));

  int sum_rslt = 7;
  batch = jsonrpc_batch_new(client);
  assert(batch);
  assert(jsonrpc_call(client, batch, "sum", json_tokener_parse("[1, 2, 4]"), sum_cb, &sum_rslt, NULL));
  assert(jsonrpc_call(client, batch, "update", json_tokener_parse("[1, 2, 3, 4, 5]"), NULL, NULL, NULL));
  assert(jsonrpc_call(client, batch, "subtract", json_tokener_parse("[42, 23]"), subtract_cb, subtract_rslts, NULL));
  assert(jsonrpc_call(client, batch, "foobar", json_tokener_parse("[]"), foobar_cb, NULL, NULL));
  assert(jsonrpc_call(client, batch, "get_data", NULL, get_data_cb, NULL, NULL));
  assert(jsonrpc_batch_send(batch));
  jsonrpc_batch_free(batch);

  assert(verto_add_idle(ctx, VERTO_EV_FLAG_NONE, on_idle));
  verto_run(ctx);
  assert(update_rslt == 15);
  assert(subtract_count == 5);
  assert(foobar_count == 2);
  assert(sum_count == 1);
  assert(get_data_count == 1);
  printf("UPDATE 1:\tOK\n");

egress:
  verto_del(client);
  verto_del(server);
  close(sockpair[0]);
  close(sockpair[1]);
  verto_free(ctx);
  return retval;
}
